#  video.py
#  This file is part of ytsub 
# 
#  Copyright (C) 2013 - Jackson Williams

#  ytsub is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ytsub is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ytsub.  If not, see <http://www.gnu.org/licenses/>.

from functools import total_ordering
from datetime import datetime
from datetime import timedelta

@total_ordering
class Vid:
    def __init__(self, author_name, playlist_item_video):
        self.id = playlist_item_video["snippet"]["resourceId"]["videoId"]
        self.title = playlist_item_video["snippet"]["title"]
        self.date = datetime.strptime(\
                playlist_item_video["snippet"]["publishedAt"],\
                "%Y-%m-%dT%H:%M:%S.000Z")
        self.author = author_name
        
    def __str__(self):
        return "Vid{id:"+self.id+ \
                ", date:"+self.date.isoformat()+ \
                ", title:"+self.title+ \
                ", author:"+self.author+"}"
        
    def __eq__(self, other):
        return self.id == other.id
        
    def __lt__(self, other):
        return self.date < other.date
        
    def __hash__(self):
        return hash(self.id)



#  batch.py
#  This file is part of ytsub 
# 
#  Copyright (C) 2013 - Jackson Williams

#  ytsub is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ytsub is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ytsub.  If not, see <http://www.gnu.org/licenses/>.


from threading import RLock
from apiclient.errors import HttpError
from copy import deepcopy
from credentials import acquire_credentials
from apiclient.discovery import build
from ytsub.video import Vid
from datetime import datetime
from datetime import timedelta


import random
import os
import Queue
import httplib2
import logging
import sys
import threading
import time
import collections

#TODO use priority queue.  In cmd-line application, would allow us to stream output (useful for using with pipes, if some other time-intensive task uses this as input).  Prioritize depth-first rather than bredth-first.  Perhaps just a bool option?  I dunno when you'd want bredth rather than depth first, though.
# NOTE the problem with this is if we wanted to sort the output.  BUT the unix philosophy says we shouldn't.  A tool like sort should do that. Which is convenient for streaming output!

#TODO Batch requests.  Ugh, just found out this exists: https://developers.google.com/api-client-library/python/guide/batch . I don't think it obviates the need for RequestThreadPool, but it does get rid of a lot of the usecases.  Basically the only time you'd need a thread pool is with query dependencies like paging and chaining.
# So to utilize batching, the first query should have everything batched together.  Then as more requests get added to the queue, we can batch them together and send them.

#TODO switch to using the preexisting list_next methods

# How many threads to use
NUM_THREADS = 3

class QueryLimit:
    '''Tells a Query when to stop paging.
    Called after each request, and once before first request.
    Guaranteed to run once per response, even if already done.
        |-> Useful to remove response items that are outside limit.'''
    def __call__(self, for_query, request, response):
        pass

class QueryLimitCount(QueryLimit):
    def __init__(self, MAX_ITEMS):
        self._MAX_ITEMS = MAX_ITEMS
        self._item_count = 0
    
    def __call__(self, for_query, request, response):
        if self._MAX_ITEMS == -1:
            return
    
        # prevent last fetch overshoot
        for_query._kwargs['maxResults'] = \
                min(int(for_query._kwargs['maxResults']), self._MAX_ITEMS - self._item_count)
        
        # first call -- still want to set maxResults
        if response is None:
            return
            
        try:
            self._item_count += len(response['items'])
        except KeyError:
            return # no items
            
        # done if item max reached
        if self._item_count == self._MAX_ITEMS:
            for_query._done = "received MAX_ITEMS items (%i)" % self._item_count

class QueryLimitAge(QueryLimit):
    def __init__(self, MAX_AGE):
        self._active = (MAX_AGE != -1)
        if not self._active:
            return
            
        self._cutoff_date = datetime.now() - timedelta(MAX_AGE+1) #days
        
    def __call__(self, for_query, request, response):
        if not self._active:
            return
        
        # Nothing to do before first response
        if response is None:
            return
        
        def within_cutoff(vid):
            v = Vid(None,vid)
            if v.date < self._cutoff_date:
                return False
            return True
        

        
        # remove videos outside of cutoff
        vids_before_filter = len(response['items'])
        response['items'][:] = [i for i in response['items'] if within_cutoff(i)]    
        # NOTE, using list slice in order to remove from response, not just make new copy        
        
        if vids_before_filter != len(response['items']):
            return
        


class Query:
    def __init__(self, request_function, kwargs, limit=QueryLimit(), name=''):
        self._request_function = request_function
        self._done = ''
        self._name = name
        
        # accept multiple limit functions -- and ensure singles are iterable
        if not isinstance(limit, collections.Iterable):
            limit = (limit,)
        self._limit = limit
        
        # adjusted down later if MAX_ITEMS less than 50
        if 'maxResults' not in kwargs:
            kwargs['maxResults'] = 50
        self._kwargs = kwargs
        
        # setup for first call to limit -- before first request
        self._last_response = (self, self.get_last_request(), None)

        # call each limit once before starting
        for l in self._limit:
            l(*self._last_response)
    
    def get_name(self):
        return self._name
    
    #TODO raise error if haven't yet called next
    def get_last_request(self):
        return {'function':self._request_function,
                'kwargs':deepcopy(self._kwargs)}
    
    def request_next_page(self, http):
        if self._done:
            raise StopIteration(self._done)
    
        if http is None:
            raise ValueError("http is None")
    
        # send REST API request and wait for response
        response = self._request_function(**self._kwargs).execute(http=http)
        self._last_response = (self, self.get_last_request(), response)
           
        # run limits
        for l in self._limit:
            l(*self._last_response)
        
        # if a limit says done, no need to check exhausted
        if self._done:
            return response 
        
        # done if server exhausted (no more items avail)
        if 'nextPageToken' not in response:
            self._done = "Server has no more items"
            return response
        
        # not done; page query kwargs
        self._kwargs["pageToken"] = response["nextPageToken"]
        return response
   

def _clear_queue(queue):
    try:
        while queue.get(block=False):
            pass
    except Queue.Empty:
        pass
    while queue.unfinished_tasks > 0:
        queue.task_done()

def _queue_not_done(queue):
    return (not queue.empty()) or (queue.unfinished_tasks is not 0)

class RequestThreadPool:
    """Creates a thread pool to process api requests.
    Used with with statement.  Call do_processing before end of with."""
    def __init__(self, authorized_http_factory, on_response_func, thread_count=NUM_THREADS):
        self._http_factory = authorized_http_factory
        self._THREAD_COUNT = thread_count
        self._on_response_func = on_response_func

        self._request_queue = Queue.Queue()
        self._response_queue = Queue.Queue()
        
        self._exit_event = threading.Event()
        self._req_threads = []
        self._resp_thread = threading.current_thread()
        
        for i in range(self._THREAD_COUNT):
            req_thread = threading.Thread(target=self._process_requests)
            req_thread.daemon = True
            self._req_threads.append(req_thread)
        
    
    def put_request(self, request):
        self._request_queue.put(request)
    
    # TODO way for client to decide to __exit__ in on_response_func?
    def do_processing(self):
        # wait until something goes wrong or done all requests and responses
        while _queue_not_done(self._request_queue) or \
              _queue_not_done(self._response_queue):
            if self._exit_event.is_set():
                break
            
            try:
                response = self._response_queue.get(timeout=0.001)
            except Queue.Empty:
                continue
            
            logging.getLogger().debug("Response received: " + str(response))
            self._on_response_func(*response)
            self._response_queue.task_done()
        
    def _process_requests(self):
        http = self._http_factory()

        while not self._exit_event.is_set():
            try:
                query = self._request_queue.get(timeout=0.001)
            except Queue.Empty:
                continue
            
            http_failures = 0
            while True: # retry executing query
                try:
                    while True: # execute query with autopaging
                        response = query.request_next_page(http)
                        self._response_queue.put((query, query.get_last_request(), response))
                        http_failures -= 1
                except StopIteration: #successfully got every page
                    self._request_queue.task_done()
                    break
                except HttpError, e:
                    if http_failures > 8: # TODO write testcase for this -- have never executed this code
                        logging.getLogger().critical("Http requests failed too many times.  Exiting. " + str(e))
                        self.__exit__(*sys.exc_info())
                        break
                    http_failures += 1
                    logging.getLogger().info("Increasing backoff, got status code: %d" % e.resp.status)
                    time.sleep((2 ** http_failures) * 0.1 * (random.random() * 0.1 + 0.95))
                except Exception, e:
                    logging.getLogger().critical("Unexpected error in process_requests thread. Exiting. " + str(e))
                    self.__exit__(*sys.exc_info())
                    break
        logging.getLogger().debug('Exiting _process_requests thread')  
                    
    
    
    def __enter__(self):
        for t in self._req_threads:
            t.start()
        
        return self
    
    
    def __exit__(self, exc_type, exc_value, traceback):
        logging.getLogger().debug('Exiting RequestThreadPool: type:{t} value:{v} traceback:{tr}'
                                  .format(t=exc_type, v=exc_value, tr=traceback))
    
        if self._exit_event.is_set():
            return #someone else is already calling this
            
        # signal all threads to join
        self._exit_event.set()

        logging.getLogger().debug('Signalled RequestThreadPool threads to exit, waiting.')        
        
        # wait for other threads to end
        for t in self._req_threads + [self._resp_thread]:
            if t is threading.current_thread():
                continue
            t.join()
        
        logging.getLogger().debug('RequestThreadPool threads exited, exiting context.')
        
        # cleanup fields set in __enter__
        self._exit_event.clear()
        _clear_queue(self._request_queue)
        _clear_queue(self._response_queue)
        
        return False #Don't supress exceptions

def get_http_factory(credentials):
    return lambda : credentials.authorize(httplib2.Http(cache=".cache"))

def batch_query(http_factory, queries):
    '''Simple interface to RequestThreadPool to return a simple list of 
    responses to queries'''
    ret = []
    
    # TODO group by for_query:list of responses
    def on_response_func(for_query, request, response):
        ret.append((for_query, request, response))
        
    
    pool = RequestThreadPool(http_factory, on_response_func)    
    with pool:
        for q in queries:
            pool.put_request(q)
        pool.do_processing()
    
    return ret



def main(argv):
    logging.basicConfig()
    logging.getLogger().setLevel("ERROR")

    CLIENT_SECRETS_FILE = "../data/client_secrets.json"
    SCOPES = "https://www.googleapis.com/auth/youtube"
    YOUTUBE_API_SERVICE_NAME = "youtube"
    YOUTUBE_API_VERSION = "v3"

    credentials = acquire_credentials(SCOPES, CLIENT_SECRETS_FILE)

    youtube = build(YOUTUBE_API_SERVICE_NAME, YOUTUBE_API_VERSION,
        http=credentials.authorize(httplib2.Http()))
    





if __name__ == "__main__":
    main(sys.argv)



#  argparse_util.py
#  This file is part of youtube-dl 
# 
#  Copyright (C) 2013 - Jackson Williams

#  youtube-dl is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  youtube-dl is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with youtube-dl.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import sys

from argparse import Action
from argparse import ArgumentError

def _ensure_value(namespace, name, value):
    if getattr(namespace, name, None) is None:
        setattr(namespace, name, value)
    return getattr(namespace, name)

class MaxCountAction(Action):
    '''Acts like _CountAction, but raises an error when max_count exceeded'''
    def __init__(self,
                 option_strings,
                 dest,
                 max_count,
                 default=None,
                 required=False,
                 help=None,
                 max_exceeded_msg='flag included too many times ({count});'
                 ' max is {max_count}'):
        assert max_count > 0, 'max must be > 0'
        
        self.max_count = max_count
        self.msg = max_exceeded_msg
        super(MaxCountAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=0,
            default=default,
            required=required,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        
        new_count = _ensure_value(namespace, self.dest, 0) + 1
        if new_count > self.max_count:
            raise ArgumentError(self,self.msg.format(count=new_count, 
                                                     max_count=self.max_count))
        setattr(namespace, self.dest, new_count)

class ListOrStdinAction(Action):
    '''This stores a list, taking input from args or stdin if no args.'''

    def __init__(self,
                 option_strings,
                 dest,
                 default=None,
                 type=None,
                 required=False,
                 help=None,
                 metavar=None):
        if default is not None and required is True:
            raise ValueError('If required, default cannot be set; this action'
                             'reads from stdin if given no parameters.')
        
        super(ListOrStdinAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            nargs='*',
            default=default,
            type=type,
            required=required,
            help=help,
            metavar=metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        items = []
        
        if values is None or len(values) is 0:
            for line in sys.stdin:
                items.append(parser._get_value(self, line))
            setattr(namespace, self.dest, items)
            return
        
        items.extend(values)
        setattr(namespace, self.dest, items)



#  main.py
#  This file is part of ytsub 
# 
#  Copyright (C) 2013 - Jackson Williams

#  ytsub is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ytsub is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ytsub.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import sys
import httplib2
import api
import logging
import fileinput
import re
import pprint

from __init__ import __version__
from apiclient.discovery import build
from credentials import acquire_credentials
from argparse_util import ListOrStdinAction
from argparse_util import MaxCountAction

__VID_REGEX = re.compile(r'^(?:(?:(?:http://)?www\.)?youtube\.com/watch\?\S*?'
                         r'v=)?([a-zA-Z0-9_-]{11})\S*$')
__LOG_LEVELS = ('CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG')

def _days(strin):
    try:
        ret = int(strin)
    except ValueError:
        raise argparse.ArgumentTypeError('Days must be an integer. Given: ' + strin)
    
    if ret < 0:
        raise argparse.ArgumentTypeError('Days must be greater than or equal to 0. Given: ' + strin)
    return ret

def _vid_count(strin):
    try:
        ret = int(strin)
    except ValueError:
        raise argparse.ArgumentTypeError('Video count must be an integer.')
    
    if ret < 0:
        raise argparse.ArgumentTypeError('Days must be greater than or equal to 0.')
    return ret

def _videoID(url_or_id):
    match = __VID_REGEX.match(url_or_id)
    if not match:
        raise ValueError('Value must be a youtube video id or the watch url.')
    return match.group(1)

def _format_vid(format_string, vid):
    return format_string.format(vid=vid.id, date=vid.date, title=vid.title, author=vid.author)

def _setup():
    SCOPES = "https://www.googleapis.com/auth/youtube"
    YOUTUBE_API_SERVICE_NAME = "youtube"
    YOUTUBE_API_VERSION = "v3"

    logging.basicConfig()

    credentials = acquire_credentials(SCOPES)

    youtube = build(YOUTUBE_API_SERVICE_NAME, YOUTUBE_API_VERSION,
        http=credentials.authorize(httplib2.Http()))
        
    return (youtube, credentials)

def _list(args):
    youtube, credentials = _setup()
    
    watched = api.get_watched_ids(youtube, credentials)
    new = api.get_sub_vids(youtube, credentials, args.max_vids_per_sub, args.age_max)
    unwatched_new = filter(lambda x: x.id not in watched, new)
    unwatched_new.sort(reverse=True)
    for v in unwatched_new:
        print _format_vid("{vid}\t{date}\t{author}\t{title}", v)

def _mark_watched(args):
    youtube, credentials = _setup()
    
    api.mark_watched(youtube, credentials, args.ids)

def main():
    parser = argparse.ArgumentParser(
                    description='Perform useful tasks on your Youtube video'
                                ' subscriptions.')

    parser.add_argument('-v', '--verbose', 
                        action=MaxCountAction, 
                        default=0, 
                        max_count=len(__LOG_LEVELS)-1, 
                                      max_exceeded_msg='flag included too many'
                                      ' times; {count} out of range of ' + 
                                      str(__LOG_LEVELS),
                        help='Increases verbosity level.  Use multiple times '
                             'to max of {maxl}.'
                             .format(maxl=len(__LOG_LEVELS)-1))
    
    parser.add_argument('--version', 
                        action='version', 
                        version='%(prog)s '+__version__)
    
    # SUBCOMMANDS
    subparsers = parser.add_subparsers(title='subcommands')
    # SUBCOMMAND list
    list_parser = subparsers.add_parser('list',
                                        description='List your new Youtube '
                                                    'subscription videos.')
    
    list_parser.add_argument('-t', '--upload-time',
                             help='Limit to videos no older than N days. If '
                                  'option present with no argument, not '
                                  'limited.',
                             type=_days,
                             default=5,
                             const=-1,
                             nargs='?',
                             metavar='N',
                             dest='age_max')
                                 
    list_parser.add_argument('-c', '--vid-count',
                             help='Limit to N videos from each subscription. '
                                  'If option present with no argument, not '
                                  'limited.',
                             type=_vid_count,
                             default=50,
                             const=-1,
                             nargs='?',
                             metavar='N',
                             dest='max_vids_per_sub')
    
    list_parser.set_defaults(func=_list)
    
    # SUBCOMMAND mark-watched
    mark_watched_parser = subparsers.add_parser('mark-watched',
                                                description='Mark video ids as'
                                                            ' watched.')
    mark_watched_parser.add_argument('ids',
                                     type=_videoID,
                                     action=ListOrStdinAction, 
                                     metavar='Video ID')
    mark_watched_parser.set_defaults(func=_mark_watched)
             
                        
                                      

    
    args = parser.parse_args()
    
    # handle verbose
    logging.getLogger().setLevel(__LOG_LEVELS[args.verbose])
    
    args.func(args)




#  setup.py
#  This file is part of ytsub 
# 
#  Copyright (C) 2013 - Jackson Williams

#  ytsub is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ytsub is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ytsub.  If not, see <http://www.gnu.org/licenses/>.

from ytsub import __version__
from setuptools import setup

setup(
    name='ytsub',
    author='Jackson Williams',
    author_email='jackson.r.williams+ytsub@gmail.com',
    version=__version__,
    packages=['ytsub'],
    license='LICENSE.txt',
    description='List your Youtube subscription videos',
    long_description=open('README.txt').read(),
    install_requires=[
        "google-api-python-client >= 1.0"
    ],
    url = 'https://github.com/jwill392/ytsub',
    entry_points = {
        'console_scripts':
            ['ytsub = ytsub.main:main']},
    package_data = {
        'ytsub':['data/client_secrets.json']
    }
)



#  credentials.py
#  This file is part of ytsub 
# 
#  Copyright (C) 2013 - Jackson Williams

#  ytsub is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ytsub is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ytsub.  If not, see <http://www.gnu.org/licenses/>.

import pkgutil
import tempfile
import oauth2client.clientsecrets as clientsecrets
import oauth2client
import sys

from os.path import expanduser
from os import pathsep
from apiclient.discovery import build
from oauth2client.file import Storage
from oauth2client.client import OAuth2WebServerFlow
from oauth2client.tools import run

def acquire_credentials(scopes):
    client_secrets_contents = pkgutil.get_data('ytsub','data/client_secrets.json')
    if client_secrets_contents is None:
        sys.exit("Missing clientsecrets.json file.")
    
    try: 
        client_type, client_info = clientsecrets.loads(client_secrets_contents)
        if client_type in [clientsecrets.TYPE_WEB, clientsecrets.TYPE_INSTALLED]: 
            flow = OAuth2WebServerFlow( 
                client_info['client_id'], 
                client_info['client_secret'], 
                scopes, 
                redirect_uri=client_info['redirect_uris'], 
                user_agent=None, 
                auth_uri=client_info['auth_uri'], 
                token_uri=client_info['token_uri']) 
   
    except clientsecrets.InvalidClientSecretsError: 
        sys.exit("Invalid client secrets file.") 

    
    storage = Storage(expanduser("~") + "/.ytsub-oauth2.json")
    credentials = storage.get()

    if credentials is None or credentials.invalid:
      credentials = run(flow, storage)

    return credentials



#  api.py
#  This file is part of ytsub 
# 
#  Copyright (C) 2013 - Jackson Williams

#  ytsub is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  ytsub is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with ytsub.  If not, see <http://www.gnu.org/licenses/>.

import ytsub.batch as batch

from video import Vid
from datetime import datetime
from datetime import timedelta

class _Channel:
    def __init__(self, channel_response):
        self.title = channel_response["snippet"]["title"]
        self.channel_id = channel_response["snippet"]["resourceId"]["channelId"]

class _UploadPlaylist:
    def __init__(self, for_channel, channel_list_response):
        self.author_name = for_channel.title
        self.playlist_id = \
                channel_list_response["items"][0]["contentDetails"]\
                ["relatedPlaylists"]["uploads"]

def get_user_playlists(youtube, credentials):
    return youtube.channels().list(
        mine=True,
        part="contentDetails"
     ).execute()['items'][0]['contentDetails']['relatedPlaylists']
    

def get_watched_ids(youtube, credentials):
    ret = []

    id_watch_history_playlist = get_user_playlists(youtube, credentials)["watchHistory"]

    query = youtube.playlistItems().list(
        playlistId=id_watch_history_playlist,
        part="id",
        maxResults=50,
        fields="items(id,kind),nextPageToken")
        
    while query is not None:
        recently_watched_response = query.execute()

        for vid in recently_watched_response["items"]:
            ret.append(vid["id"])
        
        query = youtube.playlistItems().list_next(query, recently_watched_response)
        
    return ret

def get_updated_channels(youtube, credentials):
    updated_channels = []
    
    next_page_token = ""
    while next_page_token is not None:
        sub_channels_response = youtube.subscriptions().list(
            mine=True,
            part="snippet",
            maxResults=50,
            order="unread",
            fields="items/snippet,nextPageToken"
        ).execute()
        
        for channel in sub_channels_response["items"]:
            updated_channels.append(_Channel(channel))
        
        next_page_token = sub_channels_response.get("nextPageToken")

    return updated_channels

def channel_playlists_query(youtube, credentials, channel):
    return batch.Query(youtube.channels().list, 
                 {'id':channel.channel_id,
                  'part':'contentDetails',
                  'fields':'items/contentDetails'},
                  limit=batch.QueryLimitCount(1))

def get_videos_in_playlists(youtube, credentials, playlist_list, MAX_VIDS, MAX_AGE):
    assert MAX_VIDS >= -1
    assert MAX_AGE >= -1

    playlist_queries = []
    
    for playlist in playlist_list:
        query = batch.Query(youtube.playlistItems().list,
                            {'playlistId':playlist.playlist_id,
                             'part':'snippet'},
                            limit=(batch.QueryLimitCount(MAX_VIDS),
                                   batch.QueryLimitAge(MAX_AGE)))
        query._name = playlist
        playlist_queries.append(query)
    
    resps = batch.batch_query(batch.get_http_factory(credentials), playlist_queries)
    uploaded_vids = []
    for r in resps:
        author = r[0]._name.author_name
        response = r[2]
        
        uploaded_vids.extend([Vid(author, item) for item in response['items']])
    
    return uploaded_vids

def get_sub_vids(youtube, credentials, MAX_VIDS, MAX_AGE):
    channels = get_updated_channels(youtube, credentials)
    
    playlist_queries = []
    for ch in channels:
        q = channel_playlists_query(youtube, credentials, ch)
        q._name = ch
        playlist_queries.append(q)
    
    resps = batch.batch_query(batch.get_http_factory(credentials), playlist_queries)
    upload_playlists = [_UploadPlaylist(r[0]._name, r[2]) for r in resps]
    
    return get_videos_in_playlists(youtube, credentials, upload_playlists, MAX_VIDS, MAX_AGE)

def mark_watched(youtube, credentials, vids):
    history_playlist = get_user_playlists(youtube, credentials)["watchHistory"]
    
    for vid in vids:
        youtube.playlistItems().insert(part='snippet', 
                                       body={'snippet':{'playlistId':history_playlist,
                                                        'resourceId': {'videoId':vid,
                                                                       'kind':'youtube#video'}}}).execute()
    
    
    
